/******************************************************************************
 *   "$Id:  $"
 *
 *                 Copyright (c) 2000-2002  O'ksi'D
 *
 *                      All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *      Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *      Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *      Neither the name of O'ksi'D nor the names of its contributors
 *      may be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER 
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *   Author : Jean-Marc Lienher ( http://oksid.ch )
 *
 ******************************************************************************/


#include "xd640/Xd6Std.h"
#include "xd640/Xd6HtmlPrint.h"
#include "xd640/Xd6VirtualKeyboard.h"
#include "xd640/Xd6Tabulator.h"
#include "xd640/Xd6HtmlTagA.h"
#include "xd640/Xd6HtmlToRtf.h"
#include "callbacks.h"
#include <FL/Fl.H>
#include <FL/fl_ask.H>
#include <FL/Fl_File_Chooser.H>
#include <FL/x.H>
#  ifdef MSDOS
#     include <mwin/nano-X.h>
#	  include <IO.h>
#	  include <dir.h>
#	  include <stdlib.h>
#	  include <dos.h>
//#include <iostream>
#include <fstream>
//#include <stdio.h>
#include <string.h>
extern "C" {
void toggle_text_mode();
}
extern "C" {
 int unrtfgo(int argc,char **argv); //in libunrtf.a
}
#  endif

#include <string>
using std::string;

#if 1 //MSDOS
	#define _(String) String
#else
	#define _(String) gettext((String))
#endif

extern Xd6ConfigFile *cfg;

int load_plain_text(char *f); //declaration
int load_pdf(char *f); //declaration

void cb_atexit()
{
	//if (!GUI::self) return;
#if MSDOS	
	char *loc = Xd6ConfigFile::temp(); //generates dir if not exists
	char directoryname[255];
	int rescode;
	char *delname;
	fnsplit(loc, 0, directoryname, 0, 0); //loc is always present, so always rmdir
	
	DIR *d = opendir(directoryname);
	if (d){ //if directory present
	//erase files before deleting directory
	//DIR *d = opendir(directoryname); //done above already
	struct dirent *de = NULL;
	while ((de = readdir(d))){ //need two ( to avoid warning
		if (de->d_type == DT_REG) { //standard files only
			delname=strdup(directoryname); //or both point to same location!
			delname=strcat(delname,de->d_name);
			rescode = remove(delname);
		}
	}
	closedir(d);
	
	rescode = rmdir(directoryname);
	
	}//(d)
	GrClose();
#endif	
}

void cb_new(Fl_Widget*, void*)
{
	if (!GUI::self) return;

	if (GUI::self->editor->view->frame->modified) {
		int r;
		//r = fl_ask(_("File modified. Save it ?"));
		r = fl_choice("File modified. Save it ?","No","Yes",0);
		if (r) cb_save_as(NULL, NULL);
	}
	//cb_atexit(); //replaced with code above
	GUI::self->editor->file=NULL; //do not point to old file
	GUI::self->label("FlWriter 2012                                                        ");   
	GUI::self->editor->view->frame->modified = 0;
	GUI::self->editor->view->blank();
}

void cb_open(Fl_Widget*, void*)
{
	if (GUI::self) {
		char title_cstr[100];
		const char *f;
		f = fl_file_chooser("Open ...", "*.{html,htm,HTM}", ".");
		if (f) {
            GUI::self->editor->view->frame->modified = 0;
			GUI::self->editor->view->blank(); //inserted - gp
			GUI::self->load(f);
			
			sprintf(title_cstr,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
			GUI::self->copy_label(title_cstr);
		}
	} //GUI::self
}

void cb_import(Fl_Widget*, void*)
{
	if (GUI::self) {
		const char *f;
		f = fl_file_chooser("Import RTF ...", "*.{rtf,RTF}", ".");
		if (f) {
			GUI::self->editor->view->frame->modified = 0;
			GUI::self->load_rtf(f);
		}
	}
}

void cb_save(Fl_Widget*, void*)
{
	if (GUI::self) {
		if (GUI::self->editor->file) {
			GUI::self->editor->view->frame->modified = 0;
			GUI::self->editor->view->frame->to_html(
				GUI::self->editor->file);
			//below would cause blank lines to be removed in the text
			GUI::self->load(GUI::self->editor->file);
		} else {
			cb_save_as(NULL, NULL);
		}
	}
}

void cb_save_as(Fl_Widget*, void*)
{
	if (GUI::self) {
		char title_cstr[100];
		const char *f;
		f = fl_file_chooser("Save as ...", "*.{html,htm,HTM}", 
			GUI::self->editor->file);
		if (f) {
			GUI::self->editor->view->frame->modified = 0;
			GUI::self->editor->view->frame->to_html(f);
			free(GUI::self->editor->file);
			GUI::self->editor->file = strdup(f);
			sprintf(title_cstr,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
            GUI::self->copy_label(title_cstr);
			//below would cause blank lines to be removed in the text
			//GUI::self->load(GUI::self->editor->file);
		}
	}
}

void cb_export(Fl_Widget*, void*)
{
#if MSDOS
	if (GUI::self) {
		const char *f;
	string rtfname;
	string commandstring;
	rtfname=GUI::self->editor->file;
	char ftmp[255];
	fnsplit(rtfname.c_str(), 0, 0, ftmp, 0);
	rtfname = ftmp;
	rtfname += ".rtf";

		f = fl_file_chooser("Export RTF as ...", "*.{RTF, rtf},",rtfname.c_str());
		if (f) {
			Xd6HtmlToRtf r(GUI::self->editor->view->frame);
			r.to_rtf(f);
		}
	
	//try to open file to see if generated
	//errors may have occured but still a file available to use
	int tmpfd=_open(f,O_RDONLY);
	if (tmpfd == -1){
		_close(tmpfd);
		commandstring="File could not be exported!";
	}else{
		_close(tmpfd);
		fnsplit(f, 0, 0, ftmp, 0);
		commandstring=ftmp;
		commandstring += ".rtf";
		commandstring += " file successfully created!";
	}
	fl_message(commandstring.c_str());
	}
#endif
}

void cb_export_latex_text(Fl_Widget* w, void* udata)
{
	if (!GUI::self) return;
#if MSDOS	
	const char *f;
	char tmpfile[255]="t2m2p.rtf"; //choose rare name
	string commandstring;
	string texname;

	f=tmpfile;
	Xd6HtmlToRtf r(GUI::self->editor->view->frame);
	r.to_rtf(f);
		
	texname += GUI::self->editor->file;
	fnsplit(texname.c_str(), 0, 0, tmpfile, 0);
	texname = tmpfile;
#if 0 //unrtf separate file
	if ((int)udata==0){
	texname += ".tex";
	commandstring = "unrtf --latex t2m2p.rtf >" + texname;
	} else {
	texname += ".txt";
	commandstring = "unrtf --text t2m2p.rtf >" + texname;
	}
	system(commandstring.c_str());
#else //unrtf linked in
	char *argarray[3];
	argarray[2]="t2m2p.rtf"; 
	
	if ((int)udata==0){
	texname += ".tex";
	argarray[1]="--latex";
	//commandstring = "unrtf --latex t2m2p.rtf >" + texname;
	} else {
	texname += ".txt";
	argarray[1]="--text";
	//commandstring = "unrtf --text t2m2p.rtf >" + texname;
	}
	
	FILE *funrtf = freopen(texname.c_str(), "wb", stdout);
	unrtfgo (3,argarray); //"unrtf --text t2m2p.rtf" -> 3 strings
    fclose(funrtf);
#endif //unrtf
	commandstring = "del t2m2p.rtf";

	struct ffblk ffblk_tmp;
	if (!(findfirst("t2m2p.rtf", &ffblk_tmp, FA_HIDDEN | FA_SYSTEM))){
	system(commandstring.c_str());//clear tmp file
	}

	//try to open file to see if generated
	//errors may have occured but still a file available to use
	int tmpfd=_open(texname.c_str(),O_RDONLY);
	if (tmpfd == -1){
		_close(tmpfd);
		commandstring="File could not be exported!";
	}else{
		_close(tmpfd);
		commandstring=texname;
		commandstring += " file successfully created!";
	}
	fl_message(commandstring.c_str());
#endif	
}


void cb_print(Fl_Widget* w, void*)
{
	string commandstring;
	string resultstring;
	string psname;

	if (!GUI::self) return;
#if MSDOS
    Fl_Menu_Item rclick_menu[] = {
			{"Postscript", 0, cb_ps_print,NULL},
			{"HP Deskjet", 0, cb_pcl_print,(void*)"deskjet"},
			{"HP DeskJet 500", 0, cb_pcl_print,(void*)"djet500"}, 
			{"HP LaserJet", 0, cb_pcl_print,(void*)"laserjet"},  
			{"HP LaserJet plus", 0, cb_pcl_print,(void*)"ljetplus"},  
			{"HP LaserJet IId", 0, cb_pcl_print,(void*)"ljet2p"},  
			{"HP LaserJet III", 0, cb_pcl_print,(void*)"ljet3"},  
			{"HP LaserJet 4", 0, cb_pcl_print,(void*)"ljet4"},  
			{"HP LaserJet 4d", 0, cb_pcl_print,(void*)"ljet4d"},  
			{"HP PaintJet XL printer", 0, cb_pcl_print,(void*)"pj"},  
		    {"HP PaintJet XL color printer", 0, cb_pcl_print,(void*)"pjxl"},  
			{"HP PaintJet XL300 color printer", 0, cb_pcl_print,(void*)"pjxl300"},  
		    {"HP DeskJet 500C", 0, cb_pcl_print,(void*)"cdeskjet"},  
		    {"HP DeskJet 500C 24bit", 0, cb_pcl_print,(void*)"cdjcolor"},  
		    {"HP DeskJet 500C mono", 0, cb_pcl_print,(void*)"cdjmono"},  
			{"HP DeskJet 550C", 0, cb_pcl_print,(void*)"cdj550"}, 
			{"Epson matrix printer", 0, cb_pcl_print,(void*)"epson"},  
			{"Epson hires matrix printer", 0, cb_pcl_print,(void*)"eps9high"}, 
			{"IBM matrix Proprinter", 0, cb_pcl_print,(void*)"ibmpro"}, 
			{"Canon BubbleJet BJ10e", 0, cb_pcl_print,(void*)"bj10e"}, 
			{"Canon BubbleJet BJ200", 0, cb_pcl_print,(void*)"bj200"}, 
			{"Canon BubbleJet BJC-600", 0, cb_pcl_print,(void*)"bjc600"}, 
			{"Canon BubbleJet BJC-800", 0, cb_pcl_print,(void*)"bjc800"}, 
		{ 0 }
    };
    const Fl_Menu_Item *m = rclick_menu->popup(Fl::event_x(), Fl::event_y(), 0, 0, 0);
    if ( m ) m->do_callback(w, m->user_data());
    return;
#else
	if (!GUI::self->editor->view->frame->wysiwyg) {
		fl_alert("You must set Zoom to 100 before printing.");
		return;
	};
	GUI::self->editor->view->print("mytest.ps");
#endif
}	

char* make_ps_file(Fl_Widget* w, void* filename)
{//helper function for print callbacks below
#if MSDOS
	string commandstring;
	string resultstring;
	string psname;
	char * psnameptr = new char[256];

	//int ri;
	//char ri_string[3];

	if (!GUI::self) return NULL;

	cb_save(NULL, NULL);

	if (filename == NULL){
	psname += GUI::self->editor->file;
	char f[255];
	fnsplit(psname.c_str(), 0, 0, f, 0);
	psname = f;
	psname += ".ps";
	} else { 
	psname = (char*)filename;
	}
	commandstring = "del " + psname;
	struct ffblk ffblk_tmp;
	if (!(findfirst(psname.c_str(), &ffblk_tmp, FA_HIDDEN | FA_SYSTEM))){
	system(commandstring.c_str());//clear old file first to see if no one is created
	}
	strcpy(psnameptr,psname.c_str());
	int direct = GUI::self->editor->view->print(psnameptr);

#if 0
	//commandstring = "htmldoc --webpage --header \"...\" --no-strict --quiet -f ";
	//q=quiet, w=webpage, le=left, ri=right, bodyf=bodyfont, fontsi=fontsize, no-s=no-strict, si=size
	//- not to exceed command line length
	setenv("HTMLDOC_DATA",".\\fonts\\htmldoc",0); 
	commandstring = "htmldoc --q --bodyf Arial --fontsi 12 --w --header \"...\" --no-s ";
	if (GUI::self->editor->view->landscape != 0){
	commandstring += "--la ";
	}
	if (GUI::self->editor->view->margin_left != 0){ 
	commandstring += "--le "; //below: 72 points is htmldoc default
	//ri=72 + GUI::self->editor->view->margin_left;
	ri=GUI::self->editor->view->margin_left;
	ri += 72;
	if (ri>98) ri=98; //htmldoc does not accept higher point values
	sprintf(ri_string,"%d",ri);
	commandstring += ri_string;
	} 
	commandstring += " --ri ";
	if (GUI::self->editor->view->paper_size == 1){ //letter
    ri=43 + GUI::self->editor->view->margin_right;
	sprintf(ri_string,"%d",ri);
	commandstring += ri_string;
	commandstring += " --si Letter -f ";
	} else {
	ri=20 + GUI::self->editor->view->margin_right;
	sprintf(ri_string,"%d",ri);
	commandstring += ri_string;
	commandstring += " --si A4 -f ";
	}
	commandstring +=psname + " ";
	commandstring += GUI::self->editor->file;
	system(commandstring.c_str());
#endif

	//try to open printit.ps to see if generated
	//errors may have occured but still a file available to use
	int tmpfd=_open(psname.c_str(),O_RDONLY);
	if (tmpfd == -1){
		_close(tmpfd);
		return NULL; //no file present - could be an old file though
	}else{
		_close(tmpfd);
		if (direct==0){
		return "direct";
		}else{
		return (char*)psname.c_str();
		}
	}
#endif
  return 0;
}

void cb_ps_print(Fl_Widget* w, void*)
{
	string resultstring;
	string psname;

	if (!GUI::self) return;

#if MSDOS
	char* retname=make_ps_file(NULL,NULL);
	psname=retname;
	if (retname==NULL){
	resultstring="Postscript file could not be created!";
	fl_message(resultstring.c_str());
	}else{
		if (strcmp(retname, "direct") != 0){ //no direct print - output success message
		resultstring=psname;
		resultstring += " file successfully created!";
		fl_message(resultstring.c_str());
		} //direct
	}
	return;

#else
	if (!GUI::self->editor->view->frame->wysiwyg) {
		//fl_alert(_("You must set Zoom to 100% before printing."));
		return;
	};
	GUI::self->editor->view->print();
#endif
}

void cb_pcl_print(Fl_Widget*, void* ptype){
#if MSDOS
        string commandstring;
	string resultstring;
	string psname;
	string prn_name;
	string devicename;

	devicename = (char*)ptype;

	if (!GUI::self) return;

//first make PS file

	cb_save(NULL, NULL);
	//toggle_text_mode();

	psname += GUI::self->editor->file;
	char f[255];
	fnsplit(psname.c_str(), 0, 0, f, 0);
	//psname = f;
	//psname += ".ps";
	psname = "outtmp.ps";
	prn_name = f;
	if (devicename == "pdfwrite"){
	prn_name += ".pdf";
	} else {
	prn_name += ".prn";
	}
	//commandstring = "del " + psname;
	//system(commandstring.c_str());//clear old file first to see if no one is created

	char* retname=make_ps_file(NULL,(void*)"outtmp.ps");
	psname=retname;
	if (retname==NULL){
	resultstring="Postscript file could not be created!";
	} else {

	//now call ghostscript to convert the PS file to a PRN file
//@set GS_LIB=libgs;\usr\share\htmldoc\gsfonts
//@set GS_FONTPATH=\usr\share\htmldoc\gsfonts
//	setenv("GS_LIB","libgs;\\usr\\share\\htmldoc\\gsfonts",0); //do not replace existing
//	setenv("GS_FONTPATH","\\usr\\share\\htmldoc\\gsfonts",0); //we need path in GS_LIB too!
	setenv("GS_LIB","libgs;.\\fonts\\gsfonts",0); //do not replace existing
	setenv("GS_FONTPATH",".\\fonts\\gsfonts",0); //we need path in GS_LIB too!

//gs386 -q -dNOPAUSE -sDEVICE=ljet4d -sOutputFile=%1.prn %1.ps -c quit	
	//commandstring = "dos4gw gs386 -q -dNOPAUSE -sDEVICE=";
	commandstring = "gs386 -q -dNOPAUSE -sDEVICE="; //bound dos32a to gs386 now
	commandstring += devicename + " -sOutputFile=";
	commandstring +=prn_name + " " + psname;
	commandstring += " -c quit";
	system(commandstring.c_str());

	//erase the temp ps file
	commandstring = "del outtmp.ps";
	struct ffblk ffblk_tmp;
	if (!(findfirst("outtmp.ps", &ffblk_tmp, FA_HIDDEN | FA_SYSTEM))){
	system(commandstring.c_str());//delete file
	}

	//try to open prn file to see if generated
	int tmpfd=_open(prn_name.c_str(),O_RDONLY);
	if (tmpfd == -1){
		resultstring="The print file could not be created!";
		fl_message(resultstring.c_str());
		return;
	}else{
		int retcode = 0; //default
		if (devicename.compare("pdfwrite") != 0){ //if (devicename != "pdfwrite"){
		resultstring=prn_name+" file successfully created! Print it now?";
		retcode = fl_choice(resultstring.c_str(),"NO","YES",NULL);
		} else { //pdf file success message
		resultstring=prn_name+" file successfully created!";
		fl_message(resultstring.c_str());
		}
		_close(tmpfd);	
		if (retcode==1) { //yes
			resultstring="copy /b "+prn_name;
			retcode = fl_choice("Select output device","LPT2","LPT1","PRN");
			switch (retcode) {
			case 1:
				resultstring += " lpt1:";
				break;
			case 0:
				resultstring += " lpt2:";
				break;
			case 2:
				resultstring += " prn:";
				break;
			} //switch
			system(resultstring.c_str());
			
			} //if retcode
		} //(tmpfd == -1)
	}//toggle_text_mode();
#endif
	return;
}

void cb_pdf(Fl_Widget*, void*)
{
#if MSDOS
cb_pcl_print(NULL, (void*)"pdfwrite");
#endif
}

void cb_pref(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->view->page_setup();
}

void cb_pref_font(Fl_Widget*, void*)
{
	Xd6DefaultFonts::setup();
}

void cb_exit(Fl_Widget*, void*)
{
	if (GUI::self->editor->view->frame->modified) {
		int r;
		//r = fl_ask(_("File modified. Save it ?"));
		r = fl_choice("File modified. Save it ?","No","Yes",0);
		if (r) cb_save_as(NULL, NULL);
	}
	exit(0);
}


void cb_cut(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->tool->cut();
}

void cb_copy(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->tool->copy();
}

void cb_paste(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->tool->paste();
}

void cb_find(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->view->find();
}


void cb_zoom(Fl_Widget*, void* v)
{
	int z = (int) v;
	if (!GUI::self) return;
	if (GUI::self->editor->page_width == -1) {
		GUI::self->editor->page_width =
			 GUI::self->editor->view->frame->page_width;
	}
	GUI::self->editor->view->frame->zoom(z);
	if (z != 100) {
		GUI::self->editor->view->frame->wysiwyg = 0;
	} else {
		GUI::self->editor->view->frame->page_width =
			GUI::self->editor->page_width;
		GUI::self->editor->view->frame->wysiwyg = 1;
	}

	GUI::self->editor->view->frame->measure();
	GUI::self->editor->view->frame->resize(
		GUI::self->editor->view->frame->max_width,
		GUI::self->editor->view->frame->max_height);
	GUI::self->editor->view->frame->create_pages();
	GUI::self->editor->view->redraw();
}


void cb_tabs(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	Xd6Tabulator::setup();
	GUI::self->editor->view->frame->measure();
	GUI::self->editor->view->frame->create_pages();
	GUI::self->editor->view->redraw();
}

void cb_insert(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->tool->insert_file();
}

void cb_insert_text(Fl_Widget*, void*)
{
#if MSDOS  
	if (GUI::self) {
		char *f;
		f = fl_file_chooser("Import text file...", "*.{TXT, txt, DOC, doc}", ".");
		if (f) {
			load_plain_text(f);
		}
		char title_cstr[100];
		char htmfile[255];
		fnsplit(f, 0, 0, htmfile, 0);
		strcat(htmfile,".htm");
		GUI::self->editor->view->frame->modified = 0;
		GUI::self->editor->view->blank(); //inserted - gp
		GUI::self->load(htmfile);
			
		sprintf(title_cstr,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
		GUI::self->copy_label(title_cstr);
#if 0
		//save it in FlWriter XHTML style
		GUI::self->editor->view->frame->to_html(
			GUI::self->editor->file);
		GUI::self->load(GUI::self->editor->file);
#endif
	} //GUI::self
#endif
}

void cb_insert_docx(Fl_Widget*, void*)
{
#if MSDOS  
	if (GUI::self) {
		char *f;
		f = fl_file_chooser("Import DOCX or XML file...", "*.{DOC,doc,DOCX,docx,XML,xml}", ".");
		if (f) {
			load_docx(f);
		}
		char title_cstr[100];
		char htmfile[255];
		fnsplit(f, 0, 0, htmfile, 0);
		strcat(htmfile,".htm");
		GUI::self->editor->view->frame->modified = 0;
		GUI::self->editor->view->blank(); //inserted - gp
		GUI::self->load(htmfile);
		
		//cb_save(NULL, NULL); //will convert in FlWriter style xml file
						
		sprintf(title_cstr,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
		GUI::self->copy_label(title_cstr);

	} //GUI::self
#endif
}

void cb_insert_odt(Fl_Widget*, void*)
{
#if MSDOS  
	if (GUI::self) {
		char *f;
		f = fl_file_chooser("Import ODT file...", "*.{ODT, odt}", ".");
		if (f) {
			load_odt(f);
		}
		char title_cstr[100];
		char htmfile[255];
		fnsplit(f, 0, 0, htmfile, 0);
		strcat(htmfile,".htm");
		GUI::self->editor->view->frame->modified = 0;
		GUI::self->editor->view->blank(); //inserted - gp
		GUI::self->load(htmfile);
		
		cb_save(NULL, NULL); //will convert in FlWriter style xml file
			
		sprintf(title_cstr,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
		GUI::self->copy_label(title_cstr);

	} //GUI::self
#endif
  
}

void cb_insert_pdf(Fl_Widget*, void*)
{
#if MSDOS  
	if (GUI::self) {
		char *f;
		f = fl_file_chooser("Import PDF file...", "*.{PDF, pdf}", ".");
		if (f) {
			GUI::self->load_pdf(f);
		}
		char title_cstr[100];
		char htmfile[255];
		fnsplit(f, 0, 0, htmfile, 0);
		strcat(htmfile,".htm");
		GUI::self->editor->view->frame->modified = 0;
		GUI::self->editor->view->blank(); //inserted - gp
		GUI::self->load(htmfile);
			
		sprintf(title_cstr,"FlWriter 2012 - Currently editing: %s",GUI::self->editor->file);
		GUI::self->copy_label(title_cstr);

	} //GUI::self
#endif
  
}

void cb_picture(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->tool->insert_picture->do_callback();
}

void cb_rtl_orient(Fl_Widget*, void*)
{
	if (!GUI::self) return;

	static Xd6XmlStl st;

	st.clear_flags();
	st.flags[1] = (1 << 0);

	if (st.rtl_direction == 0) {
		st.rtl_direction = 1;
	} else {
		st.rtl_direction = 0;
	}
	GUI::self->editor->view->frame->change_style(&st);
	GUI::self->editor->view->redraw();
	//Xd6HtmlToolbar::update_style(); //too difficult to call from here
}

void cb_page_break(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->tool->page_break->do_callback();
}

void cb_tag(Fl_Widget*, void*)
{
	Xd6HtmlFrame *f;
	Xd6HtmlDisplay *s;
	Xd6HtmlSegment *se;
	Xd6XmlTreeElement *e;
	Xd6XmlAttribute *a;
	const char *v;
	Xd6XmlStl st;

	st.clear_flags();
	st.a_link = 1;
	st.underline = 1;
	st.fg_color = COLOR_BLUE;
	st.flags[1] = (1 << 11) | (1 << 6) | (1 << 3);

	if (!GUI::self) return;
	v = fl_input(_("Tag Value :"));
	if (!v) return;

	GUI::self->editor->view->frame->modified = 1;
	f = GUI::self->editor->view->frame->get_cursor_frame();
	if (!f || !f->cur_chr) return;

	f->modified = 1;
	f->change_style(&st);

	e = new Xd6XmlTreeElement(NULL, 0);
	a = new Xd6XmlAttribute("href", 4);
	a->set_value(v, strlen(v));
	e->add_attribute(a);
	
	st.copy(f->cur_seg->stl);
	st.display = 1;
	st.a_link = 1;
	st.underline = 1;
	st.fg_color = COLOR_BLUE;

	e->stl = f->cur_seg->stl->get_style(&st);

	st.copy(f->cur_seg->stl);
	st.display = 0;
	st.a_link = 0;
	st.underline = 0;
	st.fg_color = COLOR_BLACK;
	f->cur_seg->stl = f->cur_seg->stl->get_style(&st);

	s = new Xd6HtmlTagA(0, e, NULL);
	f->insert_segment(s);
	se = f->cur_seg;
	se = f->cur_block->segs[se->id + 1];

	st.copy(se->stl);
	st.display = 1;
	st.a_link = 1;
	st.underline = 1;
	st.fg_color = COLOR_BLUE;

	se->stl =se->stl->get_style(&st);

	f->cur_block->measure();
	f->cur_block->create_lines();
	f->create_pages();
	GUI::self->editor->view->redraw();
	delete(e);
}

void cb_mailing(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->mailing();
	GUI::self->editor->view->redraw();
}

void cb_spell(Fl_Widget*, void*)
{
	if (!GUI::self) return;
	GUI::self->editor->view->spell();
}

void cb_lang(Fl_Widget*, void*)
{
	Xd6ConfigFile *cfg;
        Xd6ConfigFileSection *sec = NULL;
        Xd6ConfigFileGroup *grp = NULL;
        Xd6ConfigFileItem *itm = NULL;
        const char *val = NULL;
	const char *f;

        cfg = new Xd6ConfigFile("xd640", "Utilities");
        sec = cfg->get_xd640_section();
        grp = sec->add_group("spell", cfg->locale);
        itm = grp->get_create_item("dictionary", cfg->locale);
	val = itm->get_value();
	if (!val) val = "/usr/share/dict";
	f = fl_file_chooser("Dictionary ...", "*.utf8", val);

	if (f) {
		itm->set_value(f);
		cfg->write_xd640_section(sec);
	}
	delete(cfg);
}


void cb_keyboard(Fl_Widget*, void*)
{
	static Xd6VirtualKeyboard *k = NULL;
	if (!GUI::self) return;
	//if (k) delete(k);
	k = new Xd6VirtualKeyboard(GUI::self->editor->view);
	k->show();
}

void cb_about(Fl_Widget*, void*)
{
        fl_message(
#if MSDOS
_("                FlWriter 2012 - Version 1.2\n\n"
  "            ported to DOS and extended by\n\n"
  "                     Georg Potthast\n\n"
  "   based on FlWriter code by Jean-Marc Lienher\n"));
  
#else
              _("  flwriter - Copyright (c) 2000-2003 O'ksi'D\n"
                "This application is an Open Source software.\n"
                "             http://www.oksid.ch"));
#endif
}

void cb_help(Fl_Widget*, void*)
{
fl_message(_(""
"FlWriter is a text processing program using FLTK as a GUI tookit developed by Jean-Mark Lienher in 2002.\n"
"This was ported to DOS and FLTK 1.3.0 in 2012 by Georg Potthast.\n\n"
"FlWriter uses XHTML as its internal format and stores the text files in this format. \n\n"
"FlWriter features WYSIWYG. The edit window displays the text in the selected fonts. The document can \n"
"then be converted into a postscript or PCL file for printing or to a PDF document. The utility Ghostscript \n"
"is called by FlWriter for these conversions.\n"
"\n"
"There is a basic package which will not include Ghostscript and therefore cannot make PDF files or\n"
"PCL print files.\n"
"Then there is an extension package with Ghostscript, further fonts and some conversion utilities.\n"
"\nPlease consult the README.HTM file for further details.\n"
"\n\n"));

#if 0
	char buf[2048];
	char b[2048];
	
	snprintf(buf, 2048, "%s/HTML/%s/%s/index.html", 
		cfg->global_paths->doc, cfg->lang, cfg->app_name);
	if (!access(buf, 0)) {
		snprintf(b, 2048, "flspider '%s' &", buf);
	} else {
		snprintf(b, 2048, "flspider '%s/HTML/en/%s/index.html' &", 
			cfg->global_paths->doc, cfg->app_name);
	}
	system(b);
#endif
}

/*
 * Open DOS shell
 */
void cb_dos_shell(Fl_Widget*, void*)
{
#ifdef MSDOS
   toggle_text_mode();
   char *CurrWorkingDir = (char *)malloc(512);
   getcwd(CurrWorkingDir,512);

   printf("\n-> Please enter EXIT to return to FlWriter <-\n\n");       
   system("command.com");

   chdir(CurrWorkingDir);
   free(CurrWorkingDir);
   toggle_text_mode();
#endif /* MSDOS */
}


/*
 * View PS/PDF file using ghostscript
 */
void cb_view_ps_pdf(Fl_Widget*, void*)
{
	string commandstring;
	string pspdfname;

	if (GUI::self) {
		char *f;
		f = fl_file_chooser("View PS or PDF file...", "*.{PS,ps,PDF,pdf}", ".");
		if (f) {

#ifdef MSDOS

	pspdfname=f;
	
	setenv("GS_LIB","libgs;.\\fonts\\gsfonts",0); //do not replace existing
	setenv("GS_FONTPATH",".\\fonts\\gsfonts",0); //we need path in GS_LIB too!

	commandstring = "gs386 -sDEVICE=vesb -r300"; //bound dos32a to gs386 now
	commandstring +=" " + pspdfname;
	commandstring += " quit";
	
   toggle_text_mode();
   
   system(commandstring.c_str());

   toggle_text_mode();
#endif /* MSDOS */
	}//if (f)
	}//if gui::self

}



int convfsize(int fsize){
//this takes the odt/docx font sizes and converts them to FlWriter font sizes
	switch (fsize)
    {
        case 0:
			return 3; //default size
        case 1:
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7:
        case 8:
			return 1;
        case 9:
		case 10:
			return 2;
        case 11:
        case 12:
        case 13:
			return 3;
        case 14:
        case 15:
        case 16:
			return 4;
        case 17:
        case 18:
        case 19:
		case 20:
        case 21:
			return 5;
        case 22:
        case 23:
        case 24:
        case 25:
        case 26:
        case 27:
        case 28:
        case 29:
			return 6;
        case 30:
        case 31:
        case 32:
        case 33:
        case 34:
        case 35:
        case 36:
            return 7;
        default:
            return 7;
    }

return 0;
}

int load_docx(const char *f)
{
#if MSDOS 
    string commandstring;
	string resultstring;
	string docxname;
	string htmname;

	docxname = f;
	char fn[255];
	char fe[10]; //file extension
	fnsplit(docxname.c_str(), 0, 0, fn, fe);
	htmname = fn;
	htmname += ".htm";

	commandstring = "del t2m2p.xml";
	system(commandstring.c_str());//clear old file first to see if no one is created

	if (fe[1]=='x'){ //uncompressed xml file?
	//copy document.xml to t2m2p.xml now - dos copy difficult since "/" in path
	std::ifstream::pos_type size;
	char * memblock;

	std::ifstream ifile (docxname.c_str(), std::ios::in|std::ios::binary|std::ios::ate);
	std::ofstream ofile ("t2m2p.xml", std::ios::out|std::ios::binary);
	if (ifile.is_open())
	{
    size = ifile.tellg();
    memblock = new char [size];
    ifile.seekg (0, std::ios::beg);
    ifile.read (memblock, size);
	ofile.write (memblock, size);
	ifile.close();
	ofile.close();
    delete[] memblock;
	} //ifile.is_open
	}
	else{
	commandstring = "unzip32 -p ";
	commandstring += docxname;
	commandstring += " word/document.xml >t2m2p.xml";
	system(commandstring.c_str()); //extract content.xml file
	} //if not uncompressed xml file

	//try to open xml file to see if generated
	int tmpfd=_open("t2m2p.xml",O_RDONLY);
	if (tmpfd == -1){
		resultstring="The XML content could not be extracted!";
		fl_message(resultstring.c_str());
		return 0;
	}else{
		_close(tmpfd);	
	}
	
	//now convert xml to htm file
	string outchunk;
	string tmpchunk;
	int endtext_flag=1;
	int boldflag=0;
	int italicflag=0;
	int underlineflag=0;
	int fontflag=0;
	int fsize=0;
	char fontsize[5];
	char stylevalue[4];
	
	std::fstream ifs("t2m2p.xml",std::ios::in | std::ios::binary);
	std::fstream ofs(htmname.c_str(),std::ios::out | std::ios::binary | std::ios::trunc);

	while (!ifs.eof()){
	getline(ifs,outchunk,'>');	//read up to '>'
	if (!ifs.eof()){			//if not eof
		outchunk += ">";		//add > again that was removed by getline above
		
		//find paragraph end
		size_t start_pos = outchunk.find("</w:p>"); //find end of paragraph
		if(start_pos == std::string::npos){;} //if end of paragraph found
		else {        
			if (endtext_flag ==1){ //if there had been text before, no empty line
				if (boldflag==1){
				outchunk.insert(start_pos, "</strong><div></div>"); //add html paragraph
				boldflag=0;
				}else{
				outchunk.insert(start_pos, "<div></div>"); //add html paragraph
				}
				if (italicflag==1) outchunk.insert(start_pos, "</em>"); 
				italicflag=0;
				if (underlineflag==1) outchunk.insert(start_pos, "</u>"); 
				underlineflag=0;
				if (fontflag==1) outchunk.insert(start_pos, "</font>"); 
				fontflag=0;

			endtext_flag=0;
			} else{ //empty line
			outchunk.insert(start_pos, "<div>&nbsp;</div>"); //add html paragraph
			}
			endtext_flag=0;
		} //second ifs.eof()								//for FlWriter


		//find text start <w:t>
		start_pos = outchunk.find("<w:t>"); 
		if(start_pos == std::string::npos){;} 
		else {        
			if (boldflag==1) outchunk.insert(start_pos, "<strong>"); 
			if (italicflag==1) outchunk.insert(start_pos, "<em>"); 
			if (underlineflag==1) outchunk.insert(start_pos, "<u>"); 
			if (fontflag==1) {
				tmpchunk += "<font size=\"";
				sprintf(fontsize,"%d",fsize);
				tmpchunk += fontsize;	
				tmpchunk +=	"\">";
				fontflag=1; 
				outchunk.insert(start_pos, tmpchunk);
			}
		} //else

		//find text start <w:t xml...
		start_pos = outchunk.find("<w:t "); //not to confuse with w:tab!
		if(start_pos == std::string::npos){;} 
		else {        
			if (boldflag==1) {
				start_pos = outchunk.find(">"); //find end of this long tag 
				outchunk.replace(start_pos, 1, "><strong>"); 
				start_pos = start_pos+1;
			} 
			if (italicflag==1) outchunk.insert(start_pos, "<em>"); 
			if (underlineflag==1) outchunk.insert(start_pos, "<u>"); 
			if (fontflag==1) {
				tmpchunk += "<font size=\"";
				sprintf(fontsize,"%d",fsize);
				tmpchunk += fontsize;	
				tmpchunk +=	"\">";
				fontflag=1; 
				outchunk.insert(start_pos, tmpchunk);
			}
		} //else 

		//find text end
		start_pos = outchunk.find("</w:t>"); 
		if(start_pos == std::string::npos){;} 
		else {        
			endtext_flag=1;
			if (boldflag==1) outchunk.insert(start_pos,"</strong>"); 
			boldflag=0;
			if (italicflag==1) outchunk.insert(start_pos, "</em>"); 
			italicflag=0;
			if (underlineflag==1) outchunk.insert(start_pos, "</u>"); 
			underlineflag=0;
			if (fontflag==1) outchunk.insert(start_pos, "</font>"); 
			fontflag=0;
		} 

		//find pPr end - to turn off bold
		start_pos = outchunk.find("</w:pPr>"); 
		if(start_pos == std::string::npos){;} 
		else {        //this turns bold off too
			if (boldflag==1) outchunk.insert(start_pos, "</strong>"); 
			boldflag=0;
			if (italicflag==1) outchunk.insert(start_pos, "</em>"); 
			italicflag=0;
			if (underlineflag==1) outchunk.insert(start_pos, "</u>"); 
			underlineflag=0;
			if (fontflag==1) outchunk.insert(start_pos, "</font>"); 
			fontflag=0;
		} 

		//find bold
		start_pos = outchunk.find("<w:b/>"); 
		if(start_pos == std::string::npos){
			start_pos = outchunk.find("<w:b />");
			if(start_pos == std::string::npos){;}
			else {boldflag=1;}
		} else { boldflag=1; }

		//find italic
		start_pos = outchunk.find("<w:i/>"); 
		if(start_pos == std::string::npos){
			start_pos = outchunk.find("<w:i />");
			if(start_pos == std::string::npos){;}
			else {italicflag=1;}
		} else { italicflag=1;} 

		//find underline
		start_pos = outchunk.find("<w:u/>"); 
		if(start_pos == std::string::npos){
			start_pos = outchunk.find("<w:u ");
			if(start_pos == std::string::npos){;}
			else {underlineflag=1;}
		} else { underlineflag=1; } 

		//find font
		start_pos = outchunk.find("<w:sz w:val="); 
		if(!(start_pos == std::string::npos)){
			stylevalue[0]=outchunk[start_pos+13];
			stylevalue[1]=outchunk[start_pos+14];
			stylevalue[2]='\0';
			fsize=atoi(stylevalue); 
			fsize=convfsize(fsize/2); 
			fontflag=1;
		} 

		//find TAB
		start_pos = outchunk.find("<w:tab/>"); //find end of paragraph
		if(start_pos == std::string::npos){;} //if end of paragraph found
		else {
			outchunk.replace(start_pos, 9, "<w:tab/><span style=\"mso-tab-count:1\"/>"); 
		} //second ifs.eof()								//add html TAB for FlWriter

	} //first ifs.eof()
	ofs << outchunk; 
	//ofs << outchunk << std::endl; //test output
	}

	ifs.close();
	ofs.close();

#endif //MSDOS
	return 0;
}

int load_odt(const char *f)
{
#if MSDOS 
    string commandstring;
	string resultstring;
	string odtname;
	string htmname;

	odtname = f;
	char fn[255];
	fnsplit(odtname.c_str(), 0, 0, fn, 0);
	htmname = fn;
	htmname += ".htm";

	commandstring = "del t2m2p.xml";
	system(commandstring.c_str());//clear old file first to see if no one is created

	commandstring = "unzip32 -p ";
	commandstring += odtname;
	commandstring += " content.xml >t2m2p.xml";
	system(commandstring.c_str()); //extract content.xml file

	//try to open xml file to see if generated
	int tmpfd=_open("t2m2p.xml",O_RDONLY);
	if (tmpfd == -1){
		resultstring="The ODT content could not be extracted!";
		fl_message(resultstring.c_str());
		return 0;
	}else{
		_close(tmpfd);	
	}

	//now convert xml to htm file
	string outchunk;
	string tmpchunk;
	char tmpsize[5];
	int headerremoved=0;
	int initline=1;
	size_t start_pos;
	int boldflag=0;
	int italicflag=0;
	int underlineflag=0;
	int fontflag=0;
	char stylevalue[4];
	int snt=0;
	int snp=0;
	struct odt_styles otstyle[200];
	struct odt_styles opstyle[200];
	//int cstyle=0; //currentstyle
	
	std::fstream ifs("t2m2p.xml",std::ios::in | std::ios::binary);
	std::fstream ofs(htmname.c_str(),std::ios::out | std::ios::binary | std::ios::trunc);

	while (!ifs.eof()){
	//for (int i=0;i<5;i++){
	getline(ifs,outchunk,'>');	//read up to '>'
	if (!ifs.eof()){			//if not eof
		outchunk += ">";		//add > again that was removed by getline above
		if (initline==1){
			ofs << outchunk; //<?xml...
			initline=0;
		}
		if (headerremoved==0){
		  //read out style:name="Tx"
		  start_pos = outchunk.find("style:style");
		  if(!(start_pos == std::string::npos)) { //style:style
			snt=0;
			snp=0;
			start_pos = outchunk.find("style:name=\"T");
			if(!(start_pos == std::string::npos)) { //T
				//first read the Tx number of this style
				stylevalue[0]=outchunk[start_pos+13];
				stylevalue[1]=outchunk[start_pos+14];
				stylevalue[2]=outchunk[start_pos+15];
				stylevalue[3]='\0';
				snt=atoi(stylevalue);
			} else { // "T" now check "P"
			start_pos = outchunk.find("style:name=\"P");
			if(!(start_pos == std::string::npos)) {
				//first read the Px number of this style
				stylevalue[0]=outchunk[start_pos+13];
				stylevalue[1]=outchunk[start_pos+14];
				stylevalue[2]=outchunk[start_pos+15];
				stylevalue[3]='\0';
				snp=atoi(stylevalue);
			} //P
			} //T
		  } //style:style
		
		if (snp==0) {
		//this follows the T tag so snt is correct
		start_pos = outchunk.find("<style:text-properties"); 
		if(!(start_pos == std::string::npos)) {

		//clear first!	
		otstyle[snt].bold=0;
		otstyle[snt].italic=0;
		otstyle[snt].underline=0;
		otstyle[snt].fontsize=12;

		//find bold
		start_pos = outchunk.find("font-weight=\"bold"); 
		if(!(start_pos == std::string::npos))
			otstyle[snt].bold=1;

		//find italic
		start_pos = outchunk.find("font-style=\"italic"); 
		if(!(start_pos == std::string::npos))
			otstyle[snt].italic=1;

		//find underline
		start_pos = outchunk.find("style:text-underline");
		if(!(start_pos == std::string::npos))
			otstyle[snt].underline=1;

		//find fontsize
		start_pos = outchunk.find("font-size=\"");
		if(!(start_pos == std::string::npos)){
				stylevalue[0]=outchunk[start_pos+11];
				stylevalue[1]=outchunk[start_pos+12];
				stylevalue[2]='\0';
				otstyle[snt].fontsize=convfsize(atoi(stylevalue)); //atoi(stylevalue);
		}
		}// text-properties

		} else { //snp/snt
		//this follows the P tag so snp is correct
		start_pos = outchunk.find("<style:text-properties"); 
		if(!(start_pos == std::string::npos)) {

		//clear first!	
		opstyle[snp].bold=0;
		opstyle[snp].italic=0;
		opstyle[snp].underline=0;
		opstyle[snp].fontsize=12;

		//find bold
		start_pos = outchunk.find("font-weight=\"bold"); 
		if(!(start_pos == std::string::npos))
			opstyle[snp].bold=1;

		//find italic
		start_pos = outchunk.find("font-style=\"italic"); 
		if(!(start_pos == std::string::npos))
			opstyle[snp].italic=1;

		//find underline
		start_pos = outchunk.find("style:text-underline");
		if(!(start_pos == std::string::npos))
			opstyle[snp].underline=1;

		//find fontsize
		start_pos = outchunk.find("font-size=\"");
		if(!(start_pos == std::string::npos)){
				stylevalue[0]=outchunk[start_pos+11];
				stylevalue[1]=outchunk[start_pos+12];
				stylevalue[2]='\0';
				opstyle[snp].fontsize=convfsize(atoi(stylevalue));
		}
		}// text-properties
	
		} //snp/snt

		start_pos = outchunk.find("<text:p"); //find start of first paragraph
		if(!(start_pos == std::string::npos)) 
		headerremoved=1;	
		} //if headerremoved==0

		if (headerremoved==1) { //header was removed
		start_pos = outchunk.find("</text:p>"); //find end of paragraph
		if(!(start_pos == std::string::npos)) //if end of paragraph found
					outchunk.insert(start_pos, "<div></div>"); //add html paragraph for FlWriter
		
		//find empty line
		start_pos = outchunk.find("<text:p"); 
		if (!(start_pos == std::string::npos)){ //if <text:p found
			start_pos = outchunk.find("\"/>"); 
			if(!(start_pos == std::string::npos)){ //if found
			start_pos = outchunk.find("<text:p"); //restore position
			//outchunk.replace(start_pos, 36, "<text:p text:style-name=\"Standard\"/><div>&nbsp;</div>"); //add html paragraph + &nbsp;
			outchunk.insert(start_pos,"<div>&nbsp;</div>"); //add html paragraph + &nbsp;
			}
		} 

		//find span text style line
		start_pos = outchunk.find("<text:span text:style-name=\"T");
		if (!(start_pos == std::string::npos)){
				//first read the Tx number of this style
				stylevalue[0]=outchunk[start_pos+29];
				stylevalue[1]=outchunk[start_pos+30];
				stylevalue[2]=outchunk[start_pos+31];
				stylevalue[3]='\0';
				snt=atoi(stylevalue);

			tmpchunk="";
			//clear tags in reverse how they are set
			if (boldflag==1){
				tmpchunk += "</strong>";
				boldflag=0;}
			if (italicflag==1) {
				tmpchunk = tmpchunk+"</em>";
				italicflag=0; }
			if (underlineflag==1) {
				tmpchunk = tmpchunk+"</u>";
				underlineflag=0;}
			if (fontflag==1) {
				tmpchunk = tmpchunk+"</font>";
				fontflag=0;}

			if (otstyle[snt].bold==1){
				tmpchunk += "<strong>";
				boldflag=1;	}
			if (otstyle[snt].italic==1){
				tmpchunk += "<em>";
				italicflag=1; }
			if (otstyle[snt].underline==1) {
				tmpchunk += "<u>";
				underlineflag=1; }
			if (!(otstyle[snt].fontsize==0)) {
				tmpchunk += "<font size=\"";
				sprintf(tmpsize,"%d",otstyle[snt].fontsize);
				tmpchunk += tmpsize;	
				tmpchunk +=	"\">";
				fontflag=1; }
				
			outchunk.insert(start_pos, tmpchunk); 
		} //span style


		//find text:p text style paragraph
		start_pos = outchunk.find("<text:p text:style-name=\"P");
		if (!(start_pos == std::string::npos)){
				//first read the Px number of this style
				stylevalue[0]=outchunk[start_pos+26];
				stylevalue[1]=outchunk[start_pos+27];
				stylevalue[2]=outchunk[start_pos+28];
				stylevalue[3]='\0';
				snp=atoi(stylevalue);
			
			tmpchunk="";
			//clear tags in reverse how they are set
			if (boldflag==1){
				tmpchunk += "</strong>";
				boldflag=0;}
			if (italicflag==1) {
				tmpchunk = tmpchunk+"</em>";
				italicflag=0; }
			if (underlineflag==1) {
				tmpchunk = tmpchunk+"</u>";
				underlineflag=0;}
			if (fontflag==1) {
				tmpchunk = tmpchunk+"</font>";
				fontflag=0;}

			if (opstyle[snp].bold==1){
				tmpchunk += "<strong>";
				boldflag=1;	}
			if (opstyle[snp].italic==1){
				tmpchunk += "<em>";
				italicflag=1; }
			if (opstyle[snp].underline==1) {
				tmpchunk += "<u>";
				underlineflag=1; }

			if (!(opstyle[snp].fontsize==0)) {
				tmpchunk += "<font size=\"";
				sprintf(tmpsize,"%d",opstyle[snp].fontsize);
				tmpchunk += tmpsize;	
				tmpchunk +=	"\">";
				fontflag=1; }

			outchunk.insert(start_pos, tmpchunk); 

		} //paragraph style

		
		//find tab
		start_pos = outchunk.find("<text:tab/>");
		if(!(start_pos == std::string::npos)){ //if tab found
			outchunk.replace(start_pos, 9, "<text:tab/><span style=\"mso-tab-count:1\""); 
		} 			//add html TAB for FlWriter - no /> at end!

	}  //headerremoved==1
	if (headerremoved==1) ofs << outchunk; 
	//if (headerremoved==1) ofs << outchunk << endl; //test output
	//if (headerremoved==1) fprintf(stderr,"Log: %s\n",outchunk.c_str());
	} //first ifs.eof()
	}

	ifs.close();
	ofs.close();
#endif //MSDOS
	return 0;
}

void do_extended_char(char* buf,int fdoutput,short cp){

int iso;
char entity[10];
string outstring;

/* CP437 -> ISO-8859-1 */
/* Non-convertable characters are converted to a &nbsp; (0xa0) */

int table437toISO[129] =
{
  0xc7, 0xfc, 0xe9, 0xe2, 0xe4, 0xe0, 0xe5, 0xe7, 0xea, 0xeb, 0xe8, 0xef, 0xee, 0xec, 0xc4, 0xc5,
  0xc9, 0xe6, 0xc6, 0xf4, 0xf6, 0xf2, 0xfb, 0xf9, 0xff, 0xd6, 0xdc, 0xa2, 0xa3, 0xa5, 0xa0, 0xa0,
  0xe1, 0xed, 0xf3, 0xfa, 0xf1, 0xd1, 0xaa, 0xba, 0xbf, 0xa0, 0xac, 0xbd, 0xbc, 0xa1, 0xab, 0xbb,
  0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0,
  0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0,
  0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0,
  0xa0, 0xdf, 0xa0, 0xa0, 0xa0, 0xa0, 0xb5, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0,
  0xa0, 0xb1, 0xa0, 0xa0, 0xa0, 0xa0, 0xf7, 0xa0, 0xb0, 0xa0, 0xb7, 0xa0, 0xa0, 0xb2, 0xa0, 0xa0,
};

/* CP850 -> ISO-8859-1 */
/* Non-convertable characters are converted to a &nbsp; (0xa0) */

int table850toISO[129] =
{
  0xc7, 0xfc, 0xe9, 0xe2, 0xe4, 0xe0, 0xe5, 0xe7, 0xea, 0xeb, 0xe8, 0xef, 0xee, 0xec, 0xc4, 0xc5,
  0xc9, 0xe6, 0xc6, 0xf4, 0xf6, 0xf2, 0xfb, 0xf9, 0xff, 0xd6, 0xdc, 0xf8, 0xa3, 0xd8, 0xd7, 0xa0,
  0xe1, 0xed, 0xf3, 0xfa, 0xf1, 0xd1, 0xaa, 0xba, 0xbf, 0xae, 0xac, 0xbd, 0xbc, 0xa1, 0xab, 0xbb,
  0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xc1, 0xc2, 0xc0, 0xa9, 0xa0, 0xa0, 0xa0, 0xa0, 0xa2, 0xa5, 0xa0,
  0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xe3, 0xc3, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa0, 0xa4,
  0xf0, 0xd0, 0xca, 0xcb, 0xc8, 0xa0, 0xcd, 0xce, 0xcf, 0xa0, 0xa0, 0xa0, 0xa0, 0xa6, 0xcc, 0xa0,
  0xd3, 0xdf, 0xd4, 0xd2, 0xf5, 0xd5, 0xb5, 0xfe, 0xde, 0xda, 0xdb, 0xd9, 0xfd, 0xdd, 0xaf, 0xb4,
  0xad, 0xb1, 0xa0, 0xbe, 0xb6, 0xa7, 0xf7, 0xb8, 0xb0, 0xa8, 0xb7, 0xb9, 0xb3, 0xb2, 0xa0, 0xa0,
};


if (cp==437){
	iso=table437toISO[((unsigned char)*buf)-128];
} else {
	iso=table850toISO[((unsigned char)*buf)-128];

}

sprintf(entity,"%d",iso);
outstring = "&#";
outstring += entity;
outstring += ";";
write(fdoutput,outstring.c_str(),6);

}
void parse_buf(char* buf,int fdoutput,short cp){

	static int crlf=0;

	switch (*buf){
		case (char)0x0D:
			if (crlf==1) write(fdoutput,"&nbsp;",6); //empty line in code
			write(fdoutput,"</DIV\n>",7);
			break;
		case (char)0x0A:
			write(fdoutput,"<DIV\n>\n",7);
			crlf=1; break;
		case '\"':
			write(fdoutput,"&quot;",6);
			crlf=0; break;
		case '\'':
			write(fdoutput,"&apos;",6);
			crlf=0; break;
		case '&':
			write(fdoutput,"&amp;",5);
			crlf=0; break;
		case '<':
			write(fdoutput,"&lt;",4);
			crlf=0; break;
		case '>':
			write(fdoutput,"&gt;",4);
			crlf=0; break;
		case (char)0x15: //""
			write(fdoutput,"&#167;",6);
			crlf=0; break;
		case (char)0x14: //paragraph
			write(fdoutput,"&#182;",6);
			crlf=0; break;
		case (char)0x7C: // "|"
			write(fdoutput,"&#166;",6);
			crlf=0; break;

		default:
			if (strcmp(buf,"~") >0) {
				do_extended_char(buf,fdoutput,cp);
				crlf=0; break;
			}
			write(fdoutput,buf,1);
	}
}

int load_plain_text(char *f){
#if MSDOS
	char* buf= (char *)malloc(10);
	*buf='\0';
		
	int read_len;
	short cp; //codepage
	
	char outfile[255];
	fnsplit(f, 0, 0, outfile, 0);
	strcat(outfile,".htm");
	
	int fdinput = open(f, O_RDONLY|O_BINARY);
	int fdoutput = creat(outfile, S_IRUSR|S_IWUSR);

	if (fdinput < 1) return -1;
	if (fdoutput < 1) return -1;

	//read codepage
	union REGS r;
	r.x.ax = 0x6601;
	intdos(&r, &r);
	cp=r.x.bx;
	
	write(fdoutput,"<HTML><BODY><DIV\n>",18);

	read_len = read(fdinput, buf, 1);
	while (read_len > 0) {
		parse_buf(buf,fdoutput,cp);
		read_len = read(fdinput, buf, 1);
	}

	write(fdoutput,"</DIV\n></BODY></HTML>\n",22);

	close(fdinput);
	close(fdoutput);
#endif
	return 0;
}

/*
 *  End of : "$Id: $"
 */
