/******************************************************************************
 *   "$Id:  $"
 *
 *                 Copyright (c) 2000  O'ksi'D
 *
 *                      All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *      Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *      Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *      Neither the name of O'ksi'D nor the names of its contributors
 *      may be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER 
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *   Author : Jean-Marc Lienher ( http://oksid.ch )
 *
 ******************************************************************************/


#include "xd640/Xd6Std.h"
#include "gui.h"
#include "xd640/Xd6HtmlTagTable.h"
#include "xd640/Xd6SpellChoice.h"
#include <FL/fl_ask.H>
#include <FL/fl_draw.H>
#include "callbacks.h"
#include <FL/Fl_File_Chooser.H>

#include <FL/Fl_Double_Window.H>

#  ifdef MSDOS
#include <dir.h>

extern "C" {
 int unrtfgo(int argc,char **argv); //in libunrtf.a
}

#  endif

#include <string>
using std::string;

#if 1 //MSDOS
	#define _(String) String
#else
	#define _(String) gettext((String))
#endif

GUI *GUI::self = NULL;
static int text_engine(int e, Xd6HtmlFrame *frame, Xd6HtmlBlock *block,
	Xd6HtmlLine *line, Xd6HtmlSegment *seg, char *text);

GUI::GUI(int W, int H) : Fl_Double_Window(W, H)
{
        //mnu = NULL;
        mnu_bar = NULL;
        editor = NULL;
        stat_bar = NULL;
        cfg = NULL;
        cfg_sec = NULL;
        font = (Fl_Font) 0;
        size = 16;
	self = this;
	Xd6HtmlFrame::engine = &engine;
	Xd6HtmlFrame::text_engine = &text_engine;
	atexit(cb_atexit);
	fl_set_status(50, H - 25, 100, 25);
}

GUI::~GUI()
{
	self = NULL;
	delete(mnu_bar);
	delete(mnu);
	delete(editor);
	delete(stat_bar);
}

Fl_Menu_Item menuitems[] = {
{ "&File", 0, 0, 0, FL_SUBMENU },
{ "&New File", 0, (Fl_Callback *)cb_new },
{ "&Open File...", FL_COMMAND + 'o', (Fl_Callback *)cb_open },
{ "&Save File", FL_COMMAND + 's', (Fl_Callback *)cb_save },
{ "Save File &As...", 0, (Fl_Callback *)cb_save_as, 0, FL_MENU_DIVIDER },
{ "Import...", 0, 0, 0, FL_SUBMENU },
#if MSDOS
	{ "Import RTF", 0, (Fl_Callback*)cb_import, 0, 0},
	{ "Import MS DOCX/XML", 0, (Fl_Callback*)cb_insert_docx, 0, 0},
	{ "Import ODT", 0, (Fl_Callback*)cb_insert_odt, 0, 0},
#endif
	{ "Import HTML File", 0, (Fl_Callback*)cb_insert, 0, 0},
#if MSDOS
	{ "Import TEXT File", 0, (Fl_Callback*)cb_insert_text, 0, 0},
	{ "Import PDF File", 0, (Fl_Callback*)cb_insert_pdf, 0, FL_MENU_DIVIDER},
#endif
	{ 0 },
{ "Export...", 0, 0, 0, FL_SUBMENU | FL_MENU_DIVIDER},
	{ "Export RTF", 0, (Fl_Callback*)cb_export, 0, 0},
#if MSDOS
	{ "Export Latex", 0, (Fl_Callback*)cb_export_latex_text, (void*)0, 0},
	{ "Export Text", 0, (Fl_Callback*)cb_export_latex_text, (void*)1, 0},
#endif
	{ 0 },
#if MSDOS
{ "Print...", 0, 0, 0, FL_SUBMENU | FL_MENU_DIVIDER },
{ "Postscript", 0, (Fl_Callback*)cb_ps_print,0, FL_MENU_DIVIDER},
{ "HP Deskjet", 0, (Fl_Callback*)cb_pcl_print,(void*)"deskjet", 0},
{ "HP DeskJet 500", 0, (Fl_Callback*)cb_pcl_print,(void*)"djet500", 0}, 
{ "HP LaserJet", 0, (Fl_Callback*)cb_pcl_print,(void*)"laserjet", 0},  
{ "HP LaserJet plus", 0, (Fl_Callback*)cb_pcl_print,(void*)"ljetplus", 0},
{ "HP LaserJet IId", 0, (Fl_Callback*)cb_pcl_print,(void*)"ljet2p", 0},  
{ "HP LaserJet III", 0, (Fl_Callback*)cb_pcl_print,(void*)"ljet3", 0},  
{ "HP LaserJet 4", 0, (Fl_Callback*)cb_pcl_print,(void*)"ljet4", 0},  
{ "HP LaserJet 4d", 0, (Fl_Callback*)cb_pcl_print,(void*)"ljet4d", 0},  
{ "HP PaintJet XL printer", 0, (Fl_Callback*)cb_pcl_print,(void*)"pj", 0},  
{ "HP PaintJet XL color printer", 0, (Fl_Callback*)cb_pcl_print,(void*)"pjxl", 0},  
{ "HP PaintJet XL300 color printer", 0, (Fl_Callback*)cb_pcl_print,(void*)"pjxl300", 0},  
{ "HP DeskJet 500C", 0, (Fl_Callback*)cb_pcl_print,(void*)"cdeskjet", 0},  
{ "HP DeskJet 500C 24bit", 0, (Fl_Callback*)cb_pcl_print,(void*)"cdjcolor", 0},  
{ "HP DeskJet 500C mono", 0, (Fl_Callback*)cb_pcl_print,(void*)"cdjmono", 0},  
{ "HP DeskJet 550C", 0, (Fl_Callback*)cb_pcl_print,(void*)"cdj550", 0}, 
{ "Epson matrix printer", 0, (Fl_Callback*)cb_pcl_print,(void*)"epson", 0},  
{ "Epson hires matrix printer", 0, (Fl_Callback*)cb_pcl_print,(void*)"eps9high", 0}, 
{ "IBM matrix Proprinter", 0, (Fl_Callback*)cb_pcl_print,(void*)"ibmpro", 0}, 
{ "Canon BubbleJet BJ10e", 0, (Fl_Callback*)cb_pcl_print,(void*)"bj10e", 0}, 
{ "Canon BubbleJet BJ200", 0, (Fl_Callback*)cb_pcl_print,(void*)"bj200", 0}, 
{ "Canon BubbleJet BJC-600", 0, (Fl_Callback*)cb_pcl_print,(void*)"bjc600", 0}, 
{ "Canon BubbleJet BJC-800", 0, (Fl_Callback*)cb_pcl_print,(void*)"bjc800", 0}, 
{ 0 },
{ "Make PDF", FL_CTRL+'d',(Fl_Callback*)cb_pcl_print,(void*)"pdfwrite", 0},
#else
{ "Print...", 0, (Fl_Callback*)cb_print, 0, 0 },
#endif
{ "Preferences", 0, 0, 0, FL_SUBMENU | FL_MENU_DIVIDER},
	{ "Page Setup", 0, (Fl_Callback*)cb_pref, 0, 0},
	{ "Font Setup", 0, (Fl_Callback*)cb_pref_font,	0, 0},
{ 0 },
#if MSDOS
{ "DOS shell", 0, (Fl_Callback*)cb_dos_shell, 0, 0},
#endif
{ "E&xit", FL_COMMAND + 'q', (Fl_Callback *)cb_exit, 0 },
{ 0 },
{ "&Edit", 0, 0, 0, FL_SUBMENU },
{ "Cut", FL_CTRL+'x', (Fl_Callback*)cb_cut, 0, 0},
{ "Copy", FL_CTRL+'c', (Fl_Callback*)cb_copy, 0, 0},
{ "Paste", FL_CTRL+'v', (Fl_Callback*)cb_paste, 0, FL_MENU_DIVIDER},
{ "Find...", FL_CTRL+'f', (Fl_Callback*)cb_find, 0,FL_MENU_DIVIDER},
{ "Zoom", 0, 0, 0,  FL_SUBMENU },
	{ "100%", 0, (Fl_Callback*)cb_zoom,(void *)100, FL_MENU_RADIO},
	{ "150%", 0, (Fl_Callback*)cb_zoom,(void *)150, FL_MENU_RADIO},
	{ "200%", 0, (Fl_Callback*)cb_zoom,(void *)200, FL_MENU_RADIO},
	{ "300%", 0, (Fl_Callback*)cb_zoom,(void *)300, FL_MENU_RADIO},
{ 0 },
{ 0 },
{ "&Tools", 0,  0, 0, FL_SUBMENU },
	{ "Tabs", 0, (Fl_Callback*)cb_tabs, 0, FL_MENU_DIVIDER},
	{ "Show PS or PDF file", 0, (Fl_Callback*)cb_view_ps_pdf, 0, 0},
#if withpicture
    { "Insert picture", 0, (Fl_Callback*)cb_picture, 0, 0},
#endif
	{ "Insert page break", FL_CTRL+'\r', (Fl_Callback*)cb_page_break, 0, FL_MENU_DIVIDER},
	{ "Virtual keyboard", 0, (Fl_Callback*)cb_keyboard,	0, 0},
	{ "Right to left orientation", 0, (Fl_Callback*)cb_rtl_orient,	0, 0},
{ 0 },
{ "&Help", 0, 0, 0, FL_SUBMENU},
	{ "About", 0, (Fl_Callback*)cb_about, 0, 0},
	{ "Online Help", 0, (Fl_Callback*)cb_help, 0, 0},
{ 0 },	
{ 0 }
};


void GUI::create_layout()
{
	
    copy_label("FlWriter 2012");

	begin();
	mnu_bar = new Fl_Menu_Bar(0, 0, 240, 25);;
	mnu_bar->copy(menuitems);
	mnu_bar->selection_color(137);

	editor = new Xd6HtmlEditor(0, 25, 241, 146);
	resizable(editor);

	stat_bar = new Fl_Output(0, 171, 241, 20);
	stat_bar->color(FL_GRAY);
	end();

    editor->tool->save->callback(cb_save);
    editor->tool->print->callback(cb_print);
#if MSDOS
    editor->tool->pdf_file->callback(cb_pdf);
#endif    
}


int GUI::handle(int e)
{
	return Fl_Double_Window::handle(e);
}

void GUI::load(const char *f)
{
	editor->file = strdup(f);
	editor->load(f);
	editor->redraw();
}

void GUI::load_rtf(const char *f)
{
#if MSDOS  
	string commandstring;
	string rtfname;
	char filename[256];
	
	sprintf(filename,"%s",f);

	struct stat s;
	stat(filename, &s);
	if (s.st_size == 0) return; //filesize=0 - lets unrtf crash

	//cb_save(NULL, NULL); //if no file loaded is crap
	char ftmp[255];
	fnsplit(f, 0, 0, ftmp, 0);
	rtfname = ftmp;
	rtfname += ".htm";

#if 0 //unrtf separate utility
	commandstring = "unrtf ";
	commandstring += f;
	commandstring += " > " + rtfname;
	system(commandstring.c_str());
#else //with unrtf linked in
	char *argarray[2];
	argarray[1]=filename; 
    FILE *funrtf = freopen(rtfname.c_str(), "wb", stdout);
	unrtfgo (2,argarray); //"unrtf test.rtf" -> 2 strings
    fclose(funrtf);
#endif //unrtf

	//fprintf(stderr,"%s\nrtf-filename:%s\n",commandstring.c_str(),f);
	load((const char*)rtfname.c_str());
#endif	//MSDOS
}

void GUI::load_pdf(const char *f)
{
#if MSDOS  
	string commandstring;
	string pdfname;

	//cb_save(NULL, NULL); //if no file loaded is crap
	char ftmp[255];
	fnsplit(f, 0, 0, ftmp, 0);
	pdfname = ftmp;
	pdfname += ".htm";

	commandstring = "pdf2html -i -nomerge -noframes -stdout ";
	commandstring += f;
	commandstring += " >" + pdfname;
	system(commandstring.c_str());
	//fprintf(stderr,"%s\npdf-filename:%s\n",commandstring.c_str(),f);
	load((const char*)pdfname.c_str());
#endif	
}


void GUI::resize(int X, int Y, int W, int H)
{
	Window c; int mx,my,cx,cy; unsigned int mask;
	Window root;
	fl_set_status(50, H - 25, 100, 25);
	if (W == w() && H == h()) {
		Fl_Double_Window::resize(X, Y, W, H);
		return;
	}
#ifdef WIN32
	Fl_Double_Window::resize(X, Y, W, H);
	Fl_Double_Window::redraw();
#else
	root = RootWindow(fl_display, fl_screen);
	XQueryPointer(fl_display,root,&root,&c,&mx,&my,&cx,&cy,&mask);
   	if ((mask & (Button1Mask|Button2Mask|Button3Mask)) == 0) {
		Fl_Double_Window::resize(X, Y, W, H);
	//	Fl_X::i(this)->wait_for_expose = 0;
		Fl_Double_Window::redraw();
	}
#endif
}

int GUI::engine(int e, Xd6HtmlFrame *frame, Xd6HtmlBlock *block,
                Xd6HtmlLine *line, Xd6HtmlDisplay *seg, char *chr)
{
	return GUI::self->rengine(e, frame, block, line, seg, chr);
}

int GUI::rengine(int e, Xd6HtmlFrame *frame, Xd6HtmlBlock *block,
                Xd6HtmlLine *line, Xd6HtmlDisplay *seg, char *chr)
{
        if (frame != editor->view->frame) return 0;
        switch (e) {
        case FL_PUSH:
                if (Fl::event_state(FL_BUTTON3)) {
			if (seg->display == DISPLAY_TABLE) {
				table_menu(frame, block, seg);
			} else if (seg->display == DISPLAY_IMG) {
				img_menu(frame, block, seg);
			}
                }
                break;
        default:
                break;
        }
        return 0;
}

static void export_table(Xd6HtmlTagTable *t)
{
	FILE *fp;
	const char *f;

	f = fl_file_chooser("Save Table as ...", "*.{html,htm,HTM}", NULL);

	fp = fopen(f, "wb");
	if (!fp) {
		fl_alert(_("Cannot open file..."));
		return;
	}

	Xd6HtmlFrame::html_header(f, fp);
	t->to_html(fp);
        fprintf(fp, "</body>\n");
        fprintf(fp, "</html>\n");

	fclose(fp);
}

void GUI::table_menu(Xd6HtmlFrame *frame, Xd6HtmlBlock *block, 
	Xd6HtmlDisplay *seg)
{
	int i;
	const Fl_Menu_Item *ret;
	mnu = new Fl_Menu_Item[5];
	for (i = 0; i < 5; i++) {
		mnu[i].text = NULL;
	}

	mnu->add(_("Delete"), 0, 0, 0, 0);
	mnu->add(_("Export"), 0, 0, 0, 0);
	ret = mnu->popup(Fl::event_x(), Fl::event_y(), _("Table Menu"));
	if (ret == mnu) {
		//if (fl_ask(_("Really delete table ?"))) {
    if (fl_choice("Really delete table ?","Yes","No",0)) {             
			Xd6XmlStl *s;
			i = seg->id;
			s = seg->stl;
			s->para.copy(s);
			s->para.display = 0;
			s = s->get_style(&s->para);

			delete(seg);
			block->segs[i] = new 
				Xd6HtmlSegment(i, strdup(""), 0, s);
			frame->sel_chr = NULL;
			frame->cur_chr = block->segs[i]->text;
			frame->cur_seg = block->segs[i];
			frame->cur_block = block;
			frame->measure();
			frame->create_pages();
		}
	} else if (ret == mnu + 1) {
		export_table((Xd6HtmlTagTable*)seg);
	}
	delete(mnu);
	redraw();
}

static int text_engine(int e, Xd6HtmlFrame *frame, Xd6HtmlBlock *block,
	Xd6HtmlLine *line, Xd6HtmlSegment *seg, char *chr)
{
	if (e != FL_PUSH) return 0;
        if (!Fl::event_state(FL_BUTTON3)) return 0;
	GUI::self->txt_menu(frame, block, seg);
	return 1;
}

void GUI::txt_menu(Xd6HtmlFrame *frame, Xd6HtmlBlock *block, 
	Xd6HtmlSegment *seg)
{
#if MSDOS
	return;
#endif
	int i;
	const Fl_Menu_Item *ret;
	mnu = new Fl_Menu_Item[5];
	for (i = 0; i < 5; i++) {
		mnu[i].text = NULL;
	}

	mnu->add(_("Spelling"), 0, 0, 0, 0);
	ret = mnu->popup(Fl::event_x(), Fl::event_y(), _("Text Menu"));
	if (ret == mnu) {
		 editor->view->spell_choice(frame, block, seg);	
	}
	redraw();
}

void GUI::img_menu(Xd6HtmlFrame *frame, Xd6HtmlBlock *block, 
	Xd6HtmlDisplay *seg)
{
	int i;
	const Fl_Menu_Item *ret;
	mnu = new Fl_Menu_Item[5];
	for (i = 0; i < 5; i++) {
		mnu[i].text = NULL;
	}

	mnu->add(_("Delete"), 0, 0, 0, 0);
	ret = mnu->popup(Fl::event_x(), Fl::event_y(), _("Image Menu"));
	if (ret == mnu) {
		Xd6XmlStl *s;
		i = seg->id;
		s = seg->stl;
		s->para.copy(s);
		s->para.display = 0;
		s = s->get_style(&s->para);
		delete(seg);
		block->segs[i] = new Xd6HtmlSegment(i, strdup(""), 0, s);
		frame->sel_chr = NULL;
		frame->cur_chr = block->segs[i]->text;
		frame->cur_seg = block->segs[i];
		frame->cur_block = block;
		frame->measure();
		frame->create_pages();
	}
	delete(mnu);
	redraw();
}


/*
 *  End of : "$Id: $"
 */

